/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <sys/stat.h>
#include <ctype.h>
#include "spamdyke.h"
#include "environment.h"
#include "log.h"
#include "search_fs.h"

/*
 * Return value:
 *   destination_buf
 */
char *canonicalize_path(char *destination_buf, int maxlen_destination_buf, char *target_path, int strlen_target_path)
  {
  char *filename_alphabet = ALPHABET_FILENAME;
  int i;

  i = 0;

  if ((destination_buf != NULL) &&
      (maxlen_destination_buf > 0))
    {
    if (target_path != NULL)
      for (i = 0; ((i < maxlen_destination_buf) && (i < strlen_target_path)); i++)
        destination_buf[i] = (strchr(filename_alphabet, target_path[i]) != NULL) ? tolower((int)target_path[i]) : REPLACEMENT_FILENAME;

    destination_buf[i] = '\0';
    }

  return(destination_buf);
  }

/*
 * Return value:
 *   NOT FOUND: 0
 *   FOUND: 1
 */
int find_path(char *filename, char *envp[], char *return_filename, int size_return_filename)
  {
  int return_value;
  int strlen_filename;
  char *path;
  char *tmp_start;
  char *tmp_end;
  char new_filename[MAX_BUF + 1];
  struct stat tmp_stat;

  return_value = 0;

  if ((filename != NULL) &&
      (return_filename != NULL) &&
      (size_return_filename > 0))
    {
    if (stat(filename, &tmp_stat) == 0)
      {
      strlen_filename = MINVAL(size_return_filename, strlen(filename));
      strncpy(return_filename, filename, strlen_filename);
      return_filename[strlen_filename] = '\0';

      return_value = 1;
      }
    else if (strchr(filename, '/') == NULL)
      {
      if (((path = find_environment_variable(envp, ENVIRONMENT_PATH, STRLEN(ENVIRONMENT_PATH))) == NULL) ||
          (path[0] == '\0'))
        path = DEFAULT_PATH;

      tmp_start = path;
      tmp_end = NULL;
      while (tmp_start != NULL)
        {
        if ((tmp_end = strchr(tmp_start, ':')) != NULL)
          {
          strlen_filename = snprintf(new_filename, MAX_BUF, "%.*s/%s", (int)(tmp_end - tmp_start), tmp_start, filename);
          tmp_start = tmp_end + 1;
          }
        else
          {
          strlen_filename = snprintf(new_filename, MAX_BUF, "%s/%s", tmp_start, filename);
          tmp_start = NULL;
          }

        if (stat(new_filename, &tmp_stat) == 0)
          {
          if (strlen_filename > size_return_filename)
            strlen_filename = size_return_filename;

          strncpy(return_filename, new_filename, strlen_filename);
          return_filename[strlen_filename] = '\0';

          return_value = 1;
          break;
          }
        }
      }
    }

  return(return_value);
  }

/*
 * Return value:
 *   ERROR: -1
 *   NOT FOUND: 0
 *   FOUND: matching line number
 */
int search_file(struct filter_settings *current_settings, char *search_filename, char *target_string, int strlen_target_string, char start_wildcard, char *start_wildcard_matches, char end_wildcard, char *end_wildcard_matches)
  {
  int return_value;
  FILE *tmp_file;
  char tmp_buf[MAX_FILE_BUF + 1];
  char *search_buf;
  int line_num;
  int i;
  char *tmp_string;
  int strlen_buf;
  int check_start;
  int check_end;
  char lower_start_wildcard;
  char lower_end_wildcard;
  char lower_target_string[MAX_BUF + 1];
  int strlen_lower_target_string;

  return_value = 0;

  if ((target_string != NULL) &&
      (strlen_target_string > 0))
    {
    if ((tmp_file = fopen(search_filename, "r")) != NULL)
      {
      line_num = 0;

      lower_start_wildcard = (start_wildcard != '\0') ? tolower((int)start_wildcard) : start_wildcard;
      lower_end_wildcard = (end_wildcard != '\0') ? tolower((int)end_wildcard) : end_wildcard;

      strlen_lower_target_string = MINVAL(MAX_BUF, strlen_target_string);
      for (i = 0; i < strlen_lower_target_string; i++)
        lower_target_string[i] = tolower((int)target_string[i]);

      while (!feof(tmp_file) &&
             (line_num < MAX_FILE_LINES))
        {
        if ((fscanf(tmp_file, "%" STRINGIFY(MAX_FILE_BUF) "[^\r\n]", tmp_buf) == 1) &&
            (tmp_buf[0] != '#') &&
            ((strlen_buf = strlen(tmp_buf)) > 0))
          {
          for (i = 0; i < strlen_buf; i++)
            tmp_buf[i] = tolower((int)tmp_buf[i]);

          check_start = 0;
          check_end = 0;

          if ((lower_end_wildcard != '\0') &&
              (tmp_buf[strlen_buf - 1] == lower_end_wildcard))
            {
            strlen_buf--;
            tmp_buf[strlen_buf] = '\0';
            check_end = 1;
            }

          if ((lower_start_wildcard != '\0') &&
              (tmp_buf[0] == lower_start_wildcard))
            {
            strlen_buf--;
            search_buf = tmp_buf + 1;
            check_start = 1;
            }
          else
            search_buf = tmp_buf;

          if (((tmp_string = strstr(lower_target_string, search_buf)) != NULL) &&
              ((check_start &&
                ((start_wildcard_matches == NULL) ||
                 (strchr(start_wildcard_matches, (tmp_string - 1)[0]) != NULL))) ||
               (!check_start &&
                (tmp_string == lower_target_string))) &&
              ((check_end &&
                ((end_wildcard_matches == NULL) ||
                 (strchr(end_wildcard_matches, tmp_string[strlen_buf]) != NULL))) ||
               (!check_end &&
                (((tmp_string - lower_target_string) + strlen_buf) == strlen_lower_target_string))))
            {
            return_value = line_num + 1;
            break;
            }
          }

        fscanf(tmp_file, "%*1[\r\n]");
        line_num++;
        }

      if (line_num == MAX_FILE_LINES)
        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_FILE_TOO_LONG "%s", MAX_FILE_LINES, search_filename);

      fclose(tmp_file);
      }
    else
      {
      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_OPEN_SEARCH "%s: %s", search_filename, strerror(errno));
      return_value = -1;
      }
    }

  return(return_value);
  }

/*
 * Return value:
 *   ERROR: -1
 *   NOT FOUND: 0
 *   FOUND: matching line number
 */
int search_tcprules_file(struct filter_settings *current_settings, char *destination, int size_destination, char *search_filename, char *envp[], char *target_ip, char *target_name, int strlen_target_name)
  {
  int return_value;
  int line_num;
  int j;
  int k;
  FILE *tmp_file;
  char tmp_buf[MAX_FILE_BUF + 1];
  char *loc_info;
  char *loc_cmd;
  char *loc_start;
  char *env_info;
  char *tmp_name;
  char ip_octets[4][7];
  char mask_octets[4][4];
  int target_ip_ints[4];
  int ip_ints[4];
  int range_ints[4];
  int mask_ints[4];
  int network_size;
  int sscanf_result;

  return_value = 0;

  if ((target_ip != NULL) &&
      (sscanf(target_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
      (sscanf(ip_octets[0], "%d", &target_ip_ints[0]) == 1) &&
      (target_ip_ints[0] >= 0) &&
      (target_ip_ints[0] <= 255) &&
      (sscanf(ip_octets[1], "%d", &target_ip_ints[1]) == 1) &&
      (target_ip_ints[1] >= 0) &&
      (target_ip_ints[1] <= 255) &&
      (sscanf(ip_octets[2], "%d", &target_ip_ints[2]) == 1) &&
      (target_ip_ints[2] >= 0) &&
      (target_ip_ints[2] <= 255) &&
      (sscanf(ip_octets[3], "%d", &target_ip_ints[3]) == 1) &&
      (target_ip_ints[3] >= 0) &&
      (target_ip_ints[3] <= 255))
    {
    if ((tmp_file = fopen(search_filename, "r")) != NULL)
      {
      line_num = 0;

      env_info = find_environment_variable(envp, ENVIRONMENT_REMOTE_INFO, STRLEN(ENVIRONMENT_REMOTE_INFO));

      while (!feof(tmp_file) &&
             (line_num < MAX_FILE_LINES))
        {
        for (j = 0; j < 4; j++)
          {
          ip_octets[j][0] = '\0';
          ip_ints[j] = 0;
          range_ints[j] = -1;
          mask_octets[j][0] = '\0';
          mask_ints[j] = 0;
          }

        if ((fscanf(tmp_file, "%" STRINGIFY(MAX_FILE_BUF) "[^\r\n]", tmp_buf) == 1) &&
            (tmp_buf[0] != '\0') &&
            (tmp_buf[0] != '#'))
          {
          if ((loc_cmd = strchr(tmp_buf, TCPRULES_ENVIRONMENT)) != NULL)
            {
            if (loc_cmd == tmp_buf)
              {
              if (destination != NULL)
                snprintf(destination, size_destination, "%s", loc_cmd + 1);

              return_value = line_num + 1;
              break;
              }
            else
              {
              loc_cmd[0] = '\0';
              loc_cmd++;
              }
            }

          if ((loc_info = strchr(tmp_buf, TCPRULES_INFO)) != NULL)
            {
            loc_info[0] = '\0';
            if ((env_info != NULL) &&
                (strcasecmp(tmp_buf, env_info) == 0))
              loc_start = loc_info + 1;
            else
              continue;
            }
          else
            loc_start = tmp_buf;

          if (loc_start[0] == TCPRULES_NAME)
            {
            for (j = 0; tmp_buf[j] != '\0'; j++)
              tmp_buf[j] = tolower((int)tmp_buf[j]);

            if ((strlen_target_name > 0) &&
                ((loc_start[1] == '\0') ||
                 (((tmp_name = strstr(target_name, loc_start + 1)) != NULL) &&
                  ((tmp_name == target_name) ||
                   (loc_start[1] == '.') ||
                   ((tmp_name - 1)[0] == '.')))))
              {
              if (destination != NULL)
                {
                if (loc_cmd != NULL)
                  snprintf(destination, size_destination, "%s", loc_cmd);
                else
                  destination[0] = '\0';
                }

              return_value = line_num + 1;
              break;
              }
            }
          else if (((sscanf_result = sscanf(loc_start, "%7[0-9-].%7[0-9-].%7[0-9-].%7[0-9-]/%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3], mask_octets[0], mask_octets[1], mask_octets[2], mask_octets[3])) > 0) &&
                   (sscanf_result <= 4))
            {
            for (j = 0; j < sscanf_result; j++)
              if ((sscanf(ip_octets[j], "%3d-%3d", &ip_ints[j], &range_ints[j]) == 0) ||
                  (ip_ints[j] < 0) ||
                  (ip_ints[j] > 255) ||
                  ((range_ints[j] == -1) &&
                   (ip_ints[j] != target_ip_ints[j])) ||
                  ((range_ints[j] >= ip_ints[j]) &&
                   (range_ints[j] <= 255) &&
                   ((ip_ints[j] > target_ip_ints[j]) ||
                    (range_ints[j] < target_ip_ints[j]))))
                break;

            if (j == sscanf_result)
              {
              if (destination != NULL)
                {
                if (loc_cmd != NULL)
                  snprintf(destination, size_destination, "%s", loc_cmd);
                else
                  destination[0] = '\0';
                }

              return_value = line_num + 1;
              break;
              }
            }
          else if (sscanf_result == 5)
            {
            for (j = 0; j < 4; j++)
              if ((sscanf(ip_octets[j], "%3d-%3d", &ip_ints[j], &range_ints[j]) == 0) ||
                  (ip_ints[j] < 0) ||
                  (ip_ints[j] > 255))
                break;

            if ((j == 4) &&
                (sscanf(mask_octets[0], "%d", &network_size) == 1) &&
                (network_size >= 0) &&
                (network_size <= 32))
              {
              for (k = 0; k < network_size; k++)
                mask_ints[k / 8] |= 0x80 >> (k % 8);

              if (((mask_ints[0] & target_ip_ints[0]) == (mask_ints[0] & ip_ints[0])) &&
                  ((mask_ints[1] & target_ip_ints[1]) == (mask_ints[1] & ip_ints[1])) &&
                  ((mask_ints[2] & target_ip_ints[2]) == (mask_ints[2] & ip_ints[2])) &&
                  ((mask_ints[3] & target_ip_ints[3]) == (mask_ints[3] & ip_ints[3])))
                {
                return_value = line_num + 1;
                break;
                }
              }
            }
          else if (sscanf_result == 8)
            {
            for (j = 0; j < 4; j++)
              if ((sscanf(ip_octets[j], "%3d-%d", &ip_ints[j], &range_ints[j]) == 0) ||
                  (ip_ints[j] < 0) ||
                  (ip_ints[j] > 255))
                break;

            if (j == 4)
              {
              for (j = 0; j < 4; j++)
                if ((sscanf(mask_octets[j], "%d", &mask_ints[j]) != 1) ||
                    (mask_ints[j] < 0) ||
                    (mask_ints[j] > 255))
                  break;

              if ((j == 4) &&
                  ((mask_ints[0] & target_ip_ints[0]) == (mask_ints[0] & ip_ints[0])) &&
                  ((mask_ints[1] & target_ip_ints[1]) == (mask_ints[1] & ip_ints[1])) &&
                  ((mask_ints[2] & target_ip_ints[2]) == (mask_ints[2] & ip_ints[2])) &&
                  ((mask_ints[3] & target_ip_ints[3]) == (mask_ints[3] & ip_ints[3])))
                {
                if (destination != NULL)
                  {
                  if (loc_cmd != NULL)
                    snprintf(destination, size_destination, "%s", loc_cmd);
                  else
                    destination[0] = '\0';
                  }

                return_value = line_num + 1;
                break;
                }
              }
            }
          }

        fscanf(tmp_file, "%*1[\r\n]");
        line_num++;
        }

      if (line_num == MAX_FILE_LINES)
        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_FILE_TOO_LONG "%s", MAX_FILE_LINES, search_filename);

      fclose(tmp_file);
      }
    else
      {
      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_OPEN_SEARCH "%s: %s", search_filename, strerror(errno));
      return_value = -1;
      }
    }

  return(return_value);
  }

/*
 * Return value:
 *   NOT FOUND: 0
 *   FOUND: 1
 */
int search_domain_directory(struct filter_settings *current_settings, char *start_directory, char *target_domain, int strlen_target_domain)
  {
  int return_value;
  int i;
  char search_path[MAX_PATH + 1];
  char *domain_delimiter[3];
  char *base_domain;
  struct stat tmp_stat;

  return_value = 0;

  if ((start_directory != NULL) &&
      (start_directory[0] != '\0') &&
      (target_domain != NULL) &&
      (strlen_target_domain > 0))
    {
    domain_delimiter[0] = NULL;
    domain_delimiter[1] = NULL;
    domain_delimiter[2] = NULL;

    for (i = strlen_target_domain - 1; i >= 0; i--)
      if (target_domain[i] == '.')
        {
        if (domain_delimiter[0] == NULL)
          domain_delimiter[0] = target_domain + i + 1;
        else if (domain_delimiter[1] == NULL)
          domain_delimiter[1] = target_domain + i + 1;
        else if (domain_delimiter[2] == NULL)
          {
          domain_delimiter[2] = target_domain + i + 1;
          break;
          }
        }

    if (domain_delimiter[2] != NULL)
      snprintf(search_path, MAX_PATH, "%s/%.*s/%c/%.*s/%.*s/%.*s", start_directory, (int)(strlen_target_domain - (domain_delimiter[0] - target_domain)), domain_delimiter[0], domain_delimiter[1][0], (int)((domain_delimiter[0] - domain_delimiter[1]) - 1), domain_delimiter[1], (int)((domain_delimiter[1] - domain_delimiter[2]) - 1), domain_delimiter[2], strlen_target_domain, target_domain);
    else if (domain_delimiter[1] != NULL)
      snprintf(search_path, MAX_PATH, "%s/%.*s/%c/%.*s/%.*s", start_directory, (int)(strlen_target_domain - (domain_delimiter[0] - target_domain)), domain_delimiter[0], domain_delimiter[1][0], (int)((domain_delimiter[0] - domain_delimiter[1]) - 1), domain_delimiter[1], strlen_target_domain, target_domain);
    else if (domain_delimiter[0] != NULL)
      snprintf(search_path, MAX_PATH, "%s/%.*s/%c/%.*s/%.*s", start_directory, (int)(strlen_target_domain - (domain_delimiter[0] - target_domain)), domain_delimiter[0], target_domain[0], (int)((domain_delimiter[0] - target_domain) - 1), target_domain, strlen_target_domain, target_domain);
    else
      snprintf(search_path, MAX_PATH, "%s/%.*s/%c/%.*s", start_directory, strlen_target_domain, target_domain, target_domain[0], strlen_target_domain, target_domain);

    SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_INFO_DOMAIN_DIR, search_path);

    if (stat(search_path, &tmp_stat) == 0)
      return_value = 1;
    else if ((base_domain = strchr(target_domain, '.')) != NULL)
      return_value = search_domain_directory(current_settings, start_directory, base_domain + 1, strlen_target_domain - ((base_domain + 1) - target_domain));
    }

  return(return_value);
  }

/*
 * Return value:
 *   ERROR: -1
 *   SUCCESS: length of content
 */
int read_file(struct filter_settings *current_settings, char *target_filename, char **return_content, int size_return_content)
  {
  int return_value;
  FILE *tmp_file;
  char file_buf[MAX_FILE_BUF + 1];
  char *tmp_char;

  return_value = 0;
  if ((target_filename != NULL) &&
      (target_filename[0] != '\0') &&
      (return_content != NULL))
    {
    if ((tmp_file = fopen(target_filename, "r")) != NULL)
      {
      if (fscanf(tmp_file, "%" STRINGIFY(MAX_FILE_BUF) "[^\r\n]", file_buf) == 1)
        {
        return_value = (size_return_content == -1) ? strlen(file_buf) : MINVAL(size_return_content, strlen(file_buf));

        if (size_return_content == -1)
          {
          if ((tmp_char = realloc(*return_content, return_value + 1)) != NULL)
            *return_content = tmp_char;
          else
            {
            SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_REALLOC, return_value);
            return_value = -1;
            }
          }

        if (return_value > 0)
          strncpy(*return_content, file_buf, return_value);
        }

      fclose(tmp_file);
      }
    else
      {
      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_OPEN "%s: %s", target_filename, strerror(errno));
      return_value = -1;
      }
    }

  return(return_value);
  }
