/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SPAMDYKE_H
#define SPAMDYKE_H

#include <getopt.h>
#include "config.h"

#ifdef HAVE_LIBSSL

#include <openssl/ssl.h>

#define VERSION_STRING                  PACKAGE_VERSION "+TLS"

#else /* HAVE_LIBSSL */

#define VERSION_STRING                  PACKAGE_VERSION

#endif /* HAVE_LIBSSL */

#define STRLEN(X)                       (sizeof(X) - 1)
#define _STRINGIFY(X)                   #X
#define STRINGIFY(X)                    _STRINGIFY(X)

#define DEFAULT_REMOTE_IP               "0.0.0.0"
#define DEFAULT_PATH                    "/bin:/usr/bin:/usr/local/bin:/var/qmail/bin:/home/vpopmail/bin"
#define DEFAULT_TIMEOUT_COMMAND_SECS    60
#define DEFAULT_TIMEOUT_CONNECTION_SECS 0
#define TIMEOUT_CHECKPASSWORD_SECS      30
#define TIMEOUT_TLS_SHUTDOWN_SECS       5
#define TIMEOUT_COMMAND_SECS            30
#define TIMEOUT_DNS_QUERY_SECS          10

#define MIN_SELECT_SECS_TIMEOUT         0
#define MIN_SELECT_USECS_TIMEOUT        500000
#define MAX_SELECT_SECS_TIMEOUT         2
#define MAX_SELECT_USECS_TIMEOUT        0
#define SELECT_SECS_NO_TIMEOUT          2
#define SELECT_USECS_NO_TIMEOUT         0

#define ENVIRONMENT_DELIMITER           '='
#define ENVIRONMENT_REMOTE_IP           "TCPREMOTEIP"
#define ENVIRONMENT_REMOTE_NAME         "TCPREMOTEHOST"
#define ENVIRONMENT_REMOTE_INFO         "TCPREMOTEINFO"
#define ENVIRONMENT_LOCAL_PORT          "TCPLOCALPORT"
#define ENVIRONMENT_PATH                "PATH"
#define ENVIRONMENT_ALLOW_RELAY         "RELAYCLIENT"
#define ENVIRONMENT_HOSTNAME_TCPSERVER  "TCPLOCALHOST"
#define ENVIRONMENT_HOSTNAME_LINUX      "HOSTNAME"
#define ENVIRONMENT_HOSTNAME            (char *[]){ ENVIRONMENT_HOSTNAME_TCPSERVER, ENVIRONMENT_HOSTNAME_LINUX, NULL }
#define STRLEN_ENVIRONMENT_HOSTNAME     (int []){ STRLEN(ENVIRONMENT_HOSTNAME_TCPSERVER), STRLEN(ENVIRONMENT_HOSTNAME_LINUX), -1 }

#define TCPRULES_INFO                   '@'
#define TCPRULES_ENVIRONMENT            ':'
#define TCPRULES_NAME                   '='
#define TCPRULES_ALLOW                  "allow"
#define TCPRULES_DENY                   "deny"
#define TCPRULES_DELIMITER              ','

#define COMMAND_LINE_SPACER             ','
#define COMMAND_LINE_TRUE               "1tTyY"
#define COMMAND_LINE_FALSE              "0fFnN"

#define STDIN_FD                        0
#define STDOUT_FD                       1
#define STDERR_FD                       2
#define CHECKPASSWORD_FD                3

#define CHAR_CR                         '\r'
#define CHAR_LF                         '\n'
#define STR_CRLF                        "\r\n"

#define MKDIR_MODE                      0700
#define CHMOD_MODE                      0600
#define DIR_CURRENT                     "."
#define DIR_PARENT                      ".."
#define DIR_DELIMITER                   '/'

#define TEST_USER_DELIMITER             ":"

#define MAX_ADDRESS                     511
#define MAX_PATH                        4095
#define MAX_BUF                         4095
#define MAX_FILE_BUF                    65535
#define MAX_FILE_LINES                  65536
#define MAX_POLICY                      100
#define MAX_RDNS                        29
#define MAX_IP                          15
#define MAX_CHECKPASSWORD               511
#define MAX_COMMAND_BUF                 4095
#define MAX_RAND_SEED                   65536
#define MAX_DNS_QUERIES                 16
#define MAX_HOSTNAME                    127
#define MAX_BUF_SOCKET                  32768
#define MAX_DNS_PACKETS                 6
#define MAX_DNS_PACKETS_PREFERRED       1
#define MAX_DNS_TYPES                   3
#define RDNS_SUFFIX                     ".in-addr.arpa"
#define MINVAL(a,b)                     ({ typeof (a) _a = (a); typeof (b) _b = (b); _a < _b ? _a : _b; })
#define MAXVAL(a,b)                     ({ typeof (a) _a = (a); typeof (b) _b = (b); _a > _b ? _a : _b; })

#define MD5_IPAD_BYTE                   0x36
#define MD5_OPAD_BYTE                   0x5C

#define PROTOCOL_NAME_UDP               "udp"

#define USAGE_LINE_WRAP                 80
#define USAGE_LINE_INDENT               "  "
#define USAGE_MESSAGE_HEADER            "spamdyke " VERSION_STRING " (C)2007 Sam Clippinger, " PACKAGE_BUGREPORT "\nhttp://www.spamdyke.org/\n\n"
#define USAGE_MESSAGE_USAGE             "USAGE: spamdyke [ OPTIONS ] [ -- ] qmail_smtpd_command [ qmail_smtpd_arguments ]\n"
#define USAGE_MESSAGE_INTEGER_RANGE     "%s must be between (or equal to) %d and %d.\n"
#define USAGE_MESSAGE_OPTIONAL_SHORT    "No spaces are allowed between '%c' and %s.\n"
#define USAGE_MESSAGE_OPTIONAL_LONG     "No spaces are allowed and an equals sign is required between %s and %s.\n"
#define USAGE_MESSAGE_ARRAY             "%s may be used multiple times.\n"
#define USAGE_MESSAGE_SINGLETON         "%s may only be used once.\n"
#define USAGE_MESSAGE_FOOTER_SHORT      "Use -h for an option summary or see README.html for complete option details.\n\n"
#define USAGE_MESSAGE_FOOTER_LONG       "\nSee README.html for a complete explanation of these options.\n"

#define FILTER_MASK_PASS                0x03
#define FILTER_FLAG_PASS                0x00
#define FILTER_FLAG_INTERCEPT           0x01
#define FILTER_FLAG_QUIT                0x02

#define FILTER_MASK_CHILD               0x04
/* CHILD_CONTINUE is implied in assignments if CHILD_QUIT is not specified. */
#define FILTER_FLAG_CHILD_CONTINUE      0x00
#define FILTER_FLAG_CHILD_QUIT          0x04

#define FILTER_MASK_AUTH                0x18
/* AUTH_NONE is implied in assignments if another flag is not specified. */
#define FILTER_FLAG_AUTH_NONE           0x00
#define FILTER_FLAG_AUTH_ADD            0x08
#define FILTER_FLAG_AUTH_CAPTURE        0x10

#define FILTER_MASK_TLS                 0x60
/* TLS_NONE is implied in assignments if another flag is not specified. */
#define FILTER_FLAG_TLS_NONE            0x00
#define FILTER_FLAG_TLS_ADD             0x20
#define FILTER_FLAG_TLS_CAPTURE         0x40
#define FILTER_FLAG_TLS_REMOVE          0x60

#define FILTER_MASK_RCPT                0x80
/* RCPT_NONE is implied in assignments if another flag is not specified. */
#define FILTER_FLAG_RCPT_NONE           0x00
#define FILTER_FLAG_RCPT_CAPTURE        0x80

#define FILTER_MASK_CHILD_RESPONSE      0x0100
/* CHILD_RESPONSE_NONE is implied in assignments if another flag is not specified. */
#define FILTER_FLAG_CHILD_RESPONSE_NEEDED       0x0000
#define FILTER_FLAG_CHILD_RESPONSE_NOT_NEEDED   0x0100

/*
 * The values of these constants are significant.  set_config_value() uses them
 * to decide if the filter action should be set by looking at the current value
 * and the new value.  If the new value is greater than the current value, the
 * filter action is set.  Otherwise, it is not.
 */
#define FILTER_DECISION_UNDECIDED               0
#define FILTER_DECISION_DO_FILTER               1
#define FILTER_DECISION_DO_NOT_FILTER           2
#define FILTER_DECISION_CONFIG_TEST             3
#define FILTER_DECISION_FORK_ERROR              4
#define FILTER_DECISION_ERROR                   5

/*
 * The values of these constants must be in ascending order.
 */
#define FILTER_GRACE_EXPIRED            -1
#define FILTER_GRACE_NONE               0
#define FILTER_GRACE_AFTER_FROM         1
#define FILTER_GRACE_AFTER_TO           2

#define GRAYLIST_NONE                   0
#define GRAYLIST_FILTER_ALL             1
#define GRAYLIST_FILTER_NONE            2

#define TLS_STATE_INACTIVE              0
#define TLS_STATE_ACTIVE_SPAMDYKE       1
#define TLS_STATE_ACTIVE_PASSTHROUGH    2

#define LOCALHOST_IP                    "127.0.0.1"
#define LOCALHOST_OCTETS                { 127, 0, 0, 1 }
#define LOCALHOST_NAME                  "localhost"
#define MISSING_LOCAL_SERVER_NAME       "unknown.server.unknown.domain"

#define ALPHABET_FILENAME               "abcdefghijklmnopqrstuvwxyz0123456789@_-.ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define REPLACEMENT_FILENAME            '_'
#define ALPHABET_BASE64                 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
#define PAD_BASE64                      '='

#define REJECT_SEVERITY_NONE            "250 "
#define REJECT_SEVERITY_HELO            "220 "
#define REJECT_SEVERITY_QUIT            "221 "
#define REJECT_SEVERITY_AUTH_SUCCESS    "235 "
#define REJECT_SEVERITY_AUTH_CHALLENGE  "334 "
#define REJECT_SEVERITY_TEMPORARY       "421 "
#define REJECT_SEVERITY_AUTH_UNKNOWN    "504 "
#define REJECT_SEVERITY_AUTH_FAILURE    "535 "
#define REJECT_SEVERITY_PERMANENT       "554 "
#define REJECT_SEVERITY_TLS_SUCCESS     "220 "
#define REJECT_SEVERITY_TLS_FAILURE     "454 "
#define STRLEN_REJECT_SEVERITY          4
#define REJECT_CRLF                     "\r\n"

#define SMTP_HELO                       "HELO"
#define SMTP_EHLO                       "EHLO"
#define SMTP_AUTH                       "AUTH"
#define SMTP_TLS                        "STARTTLS"
#define SMTP_MAIL_FROM                  "MAIL FROM"
#define SMTP_RCPT_TO                    "RCPT TO"
#define SMTP_DATA                       "DATA"
#define SMTP_DATA_END                   "."
#define SMTP_QUIT                       "QUIT"
#define SMTP_RSET                       "RSET"

#define SMTP_AUTH_UNKNOWN               -1
#define SMTP_AUTH_LOGIN                 0
#define SMTP_AUTH_PLAIN                 1
#define SMTP_AUTH_CRAM_MD5              2
#define SMTP_AUTH_TYPE_LOGIN            "LOGIN"
#define SMTP_AUTH_TYPE_PLAIN            "PLAIN"
#define SMTP_AUTH_TYPE_CRAM_MD5         "CRAM-MD5"
#define SMTP_AUTH_TYPES                 (char *[]){ SMTP_AUTH_TYPE_LOGIN, SMTP_AUTH_TYPE_PLAIN, SMTP_AUTH_TYPE_CRAM_MD5, NULL }
#define STRLEN_SMTP_AUTH_TYPES          (int []){ STRLEN(SMTP_AUTH_TYPE_LOGIN), STRLEN(SMTP_AUTH_TYPE_PLAIN), STRLEN(SMTP_AUTH_TYPE_CRAM_MD5), -1 }
#define SMTP_AUTH_ENCRYPTION            (int []){ 0, 0, 1, -1 }

#define SMTP_AUTH_LOGIN_CHALLENGE_1             "VXNlcm5hbWU6"
#define SMTP_AUTH_LOGIN_CHALLENGE_2             "UGFzc3dvcmQ6"

#define SMTP_AUTH_ORIGIN_NONE           0
#define SMTP_AUTH_ORIGIN_SPAMDYKE       1
#define SMTP_AUTH_ORIGIN_CHILD          2

#define SMTP_AUTH_STATE_UNKNOWN                 -1
#define SMTP_AUTH_STATE_NONE                    0
#define SMTP_AUTH_STATE_CMD_SEEN                1
#define SMTP_AUTH_STATE_CHALLENGE_1_SENT        2
#define SMTP_AUTH_STATE_RESPONSE_1_SEEN         3
#define SMTP_AUTH_STATE_CHALLENGE_2_SENT        4
#define SMTP_AUTH_STATE_RESPONSE_2_SEEN         5

#define SMTP_CONTINUATION               '-'
#define SMTP_STR_CONTINUATION           "-"
#define SMTP_STR_DONE                   " "
#define SMTP_EHLO_SUCCESS               "250"
#define SMTP_EHLO_AUTH                  "AUTH "
#define SMTP_EHLO_TLS                   "STARTTLS"
#define SMTP_EHLO_AUTH_INSERT_ENCRYPTION        "AUTH LOGIN PLAIN CRAM-MD5\r\n"
#define SMTP_EHLO_AUTH_INSERT_CLEAR     "AUTH LOGIN PLAIN\r\n"
#define SMTP_EHLO_TLS_INSERT            "STARTTLS\r\n"
#define SMTP_EHLO_NOTHING_INSERT        "X-NOTHING\r\n"
#define SMTP_AUTH_CHALLENGE             "334"
#define SMTP_AUTH_SUCCESS               "235"
#define SMTP_AUTH_PROMPT                "3"
#define SMTP_AUTH_FAILURE               "5"
#define SMTP_TLS_SUCCESS                "220"
#define SMTP_TLS_REGREETING             "220 ESMTP\r\n"
#define SMTP_RCPT_SUCCESS               "250"

#define CHILD_QUIT                      ".\r\nQUIT\r\n"

#define SENDER_ADDRESS_NONE             "_none"

#define SHORT_SUCCESS                   "ALLOWED"

#define ERROR_URL                       " See: "

struct rejection_data
  {
  char *reject_severity;
  char *reject_message;
  int strlen_reject_message;
  char *short_reject_message;
  };

#define SUCCESS_AUTH                    { REJECT_SEVERITY_AUTH_SUCCESS, "Proceed.", 8, SHORT_SUCCESS "_AUTHENTICATED" }
#define SUCCESS_TLS                     { REJECT_SEVERITY_TLS_SUCCESS, "Proceed.", 8, SHORT_SUCCESS "_TLS" }

#define ERROR_FAILURE_TLS               "Failed to negotiate TLS connection."
#define FAILURE_TLS                     { REJECT_SEVERITY_TLS_FAILURE, ERROR_FAILURE_TLS, STRLEN(ERROR_FAILURE_TLS), "FAILED_TLS" }

#define REJECTION_RCPT_TO               0
#define ERROR_RCPT_TO                   "Too many recipients. Try the remaining addresses again later."
#define REJECTION_DATA_RCPT_TO          { REJECT_SEVERITY_TEMPORARY, ERROR_RCPT_TO, STRLEN(ERROR_RCPT_TO), "DENIED_TOO_MANY_RECIPIENTS" }

#define REJECTION_RCPT_TO_LOCAL         1
#define ERROR_RCPT_TO_LOCAL             "Improper recipient address. Try supplying a domain name."
#define REJECTION_DATA_RCPT_TO_LOCAL    { REJECT_SEVERITY_PERMANENT, ERROR_RCPT_TO_LOCAL, STRLEN(ERROR_RCPT_TO_LOCAL), "DENIED_UNQUALIFIED_RECIPIENT" }

#define REJECTION_GRAYLISTED            2
#define ERROR_GRAYLISTED                "Your address has been graylisted. Try again later."
#define REJECTION_DATA_GRAYLISTED       { REJECT_SEVERITY_TEMPORARY, ERROR_GRAYLISTED, STRLEN(ERROR_GRAYLISTED), "DENIED_GRAYLISTED" }

#define REJECTION_RDNS_MISSING          3
#define ERROR_RDNS_MISSING              "Refused. You have no reverse DNS entry."
#define REJECTION_DATA_RDNS_MISSING     { REJECT_SEVERITY_TEMPORARY, ERROR_RDNS_MISSING, STRLEN(ERROR_RDNS_MISSING), "DENIED_RDNS_MISSING" }

#define REJECTION_RDNS_RESOLVE          4
#define ERROR_RDNS_RESOLVE              "Refused. Your reverse DNS entry does not resolve."
#define REJECTION_DATA_RDNS_RESOLVE     { REJECT_SEVERITY_TEMPORARY, ERROR_RDNS_RESOLVE, STRLEN(ERROR_RDNS_RESOLVE), "DENIED_RDNS_RESOLVE" }

#define REJECTION_IP_IN_NAME_CC         5
#define ERROR_IP_IN_NAME_CC             "Refused. Your reverse DNS entry contains your IP address and a country code."
#define REJECTION_DATA_IP_IN_NAME_CC    { REJECT_SEVERITY_PERMANENT, ERROR_IP_IN_NAME_CC, STRLEN(ERROR_IP_IN_NAME_CC), "DENIED_IP_IN_CC_RDNS" }

#define REJECTION_IP_IN_NAME            6
#define ERROR_IP_IN_NAME                "Refused. Your reverse DNS entry contains your IP address and a banned keyword."
#define REJECTION_DATA_IP_IN_NAME       { REJECT_SEVERITY_PERMANENT, ERROR_IP_IN_NAME, STRLEN(ERROR_IP_IN_NAME), "DENIED_IP_IN_RDNS" }

#define REJECTION_EARLYTALKER           7
#define ERROR_EARLYTALKER               "Refused. You are not following the SMTP protocol."
#define REJECTION_DATA_EARLYTALKER      { REJECT_SEVERITY_PERMANENT, ERROR_EARLYTALKER, STRLEN(ERROR_EARLYTALKER), "DENIED_EARLYTALKER" }

#define REJECTION_BLACKLIST_NAME        8
#define ERROR_BLACKLIST_NAME            "Refused. Your domain name is blacklisted."
#define REJECTION_DATA_BLACKLIST_NAME   { REJECT_SEVERITY_PERMANENT, ERROR_BLACKLIST_NAME, STRLEN(ERROR_BLACKLIST_NAME), "DENIED_BLACKLIST_NAME" }

#define REJECTION_BLACKLIST_IP          9
#define ERROR_BLACKLIST_IP              "Refused. Your IP address is blacklisted."
#define REJECTION_DATA_BLACKLIST_IP     { REJECT_SEVERITY_PERMANENT, ERROR_BLACKLIST_IP, STRLEN(ERROR_BLACKLIST_IP), "DENIED_BLACKLIST_IP" }

#define REJECTION_TIMEOUT               10
#define ERROR_TIMEOUT                   "Timeout. Talk faster next time."
#define REJECTION_DATA_TIMEOUT          { REJECT_SEVERITY_TEMPORARY, ERROR_TIMEOUT, STRLEN(ERROR_TIMEOUT), "TIMEOUT" }

#define REJECTION_SENDER_BLACKLISTED    11
#define ERROR_SENDER_BLACKLISTED        "Refused. Your sender address has been blacklisted."
#define REJECTION_DATA_SENDER_BLACKLISTED       { REJECT_SEVERITY_PERMANENT, ERROR_SENDER_BLACKLISTED, STRLEN(ERROR_SENDER_BLACKLISTED), "DENIED_SENDER_BLACKLISTED" }

#define REJECTION_RECIPIENT_BLACKLISTED 12
#define ERROR_RECIPIENT_BLACKLISTED             "Refused. Mail is not being accepted at this address."
#define REJECTION_DATA_RECIPIENT_BLACKLISTED    { REJECT_SEVERITY_PERMANENT, ERROR_RECIPIENT_BLACKLISTED, STRLEN(ERROR_RECIPIENT_BLACKLISTED), "DENIED_RECIPIENT_BLACKLISTED" }

#define REJECTION_SENDER_NO_MX          13
#define ERROR_SENDER_NO_MX              "Refused. The domain of your sender address has no mail exchanger (MX)."
#define REJECTION_DATA_SENDER_NO_MX     { REJECT_SEVERITY_TEMPORARY, ERROR_SENDER_NO_MX, STRLEN(ERROR_SENDER_NO_MX), "DENIED_SENDER_NO_MX" }

#define REJECTION_RBL                   14
#define ERROR_RBL                       "Refused. Your IP address is listed in the RBL at %s"
#define REJECTION_DATA_RBL              { REJECT_SEVERITY_PERMANENT, ERROR_RBL, STRLEN(ERROR_RBL) - 2, "DENIED_RBL_MATCH" }

#define REJECTION_RHSBL                 15
#define ERROR_RHSBL                     "Refused. Your domain name is listed in the RHSBL at %s"
#define REJECTION_DATA_RHSBL            { REJECT_SEVERITY_PERMANENT, ERROR_RHSBL, STRLEN(ERROR_RHSBL) - 2, "DENIED_RHSBL_MATCH" }

#define REJECTION_SMTP_AUTH_FAILURE             16
#define ERROR_SMTP_AUTH_FAILURE                 "Refused. Authentication failed."
#define REJECTION_DATA_SMTP_AUTH_FAILURE        { REJECT_SEVERITY_AUTH_FAILURE, ERROR_SMTP_AUTH_FAILURE, STRLEN(ERROR_SMTP_AUTH_FAILURE), "FAILED_AUTH" }

#define REJECTION_SMTP_AUTH_UNKNOWN             17
#define ERROR_SMTP_AUTH_UNKNOWN                 "Refused. Unknown authentication method."
#define REJECTION_DATA_SMTP_AUTH_UNKNOWN        { REJECT_SEVERITY_AUTH_UNKNOWN, ERROR_SMTP_AUTH_UNKNOWN, STRLEN(ERROR_SMTP_AUTH_UNKNOWN), "UNKNOWN_AUTH" }

#define REJECTION_ACCESS_DENIED                 18
#define ERROR_ACCESS_DENIED                     "Refused. Access is denied."
#define REJECTION_DATA_ACCESS_DENIED            { REJECT_SEVERITY_PERMANENT, ERROR_ACCESS_DENIED, STRLEN(ERROR_ACCESS_DENIED), "DENIED_ACCESS_DENIED" }

#define REJECTION_RELAYING_DENIED               19
#define ERROR_RELAYING_DENIED                   "Refused. Sending to remote addresses (relaying) is not allowed."
#define REJECTION_DATA_RELAYING_DENIED          { REJECT_SEVERITY_PERMANENT, ERROR_RELAYING_DENIED, STRLEN(ERROR_RELAYING_DENIED), "DENIED_RELAYING" }

#define REJECTION_OTHER                         20
#define ERROR_OTHER                             "Refused by other (child) process' filter."
#define REJECTION_DATA_OTHER                    { REJECT_SEVERITY_TEMPORARY, ERROR_OTHER, STRLEN(ERROR_OTHER), "DENIED_OTHER" }

#define REJECTION_ZERO_RECIPIENTS       21
#define ERROR_ZERO_RECIPIENTS           "Refused. You must specify at least one valid recipient."
#define REJECTION_DATA_ZERO_RECIPIENTS  { REJECT_SEVERITY_PERMANENT, ERROR_ZERO_RECIPIENTS, STRLEN(ERROR_ZERO_RECIPIENTS), "DENIED_ZERO_RECIPIENTS" }

#define REJECTION_DATA                  (struct rejection_data []){ \
                                        REJECTION_DATA_RCPT_TO, \
                                        REJECTION_DATA_RCPT_TO_LOCAL, \
                                        REJECTION_DATA_GRAYLISTED, \
                                        REJECTION_DATA_RDNS_MISSING, \
                                        REJECTION_DATA_RDNS_RESOLVE, \
                                        REJECTION_DATA_IP_IN_NAME_CC, \
                                        REJECTION_DATA_IP_IN_NAME, \
                                        REJECTION_DATA_EARLYTALKER, \
                                        REJECTION_DATA_BLACKLIST_NAME, \
                                        REJECTION_DATA_BLACKLIST_IP, \
                                        REJECTION_DATA_TIMEOUT, \
                                        REJECTION_DATA_SENDER_BLACKLISTED, \
                                        REJECTION_DATA_RECIPIENT_BLACKLISTED, \
                                        REJECTION_DATA_SENDER_NO_MX, \
                                        REJECTION_DATA_RBL, \
                                        REJECTION_DATA_RHSBL, \
                                        REJECTION_DATA_SMTP_AUTH_FAILURE, \
                                        REJECTION_DATA_SMTP_AUTH_UNKNOWN, \
                                        REJECTION_DATA_ACCESS_DENIED, \
                                        REJECTION_DATA_RELAYING_DENIED, \
                                        REJECTION_DATA_OTHER, \
                                        REJECTION_DATA_ZERO_RECIPIENTS \
                                        }

#define LOG_USE_STDERR                  0
#define LOG_USE_SYSLOG                  1

#define LOG_LEVEL_NONE                  0
#define LOG_LEVEL_ERROR                 1
#define LOG_LEVEL_INFO                  2
#define LOG_LEVEL_DEBUG                 3
#define LOG_LEVEL_EXCESSIVE             4

#define SPAMDYKE_LOG_NONE(CURRENT_SETTINGS,...)         ({ if (((CURRENT_SETTINGS) == NULL) || ((CURRENT_SETTINGS)->log_level >= LOG_LEVEL_NONE)) spamdyke_log(CURRENT_SETTINGS,LOG_LEVEL_NONE,__VA_ARGS__); })
#define SPAMDYKE_LOG_ERROR(CURRENT_SETTINGS,...)        ({ if (((CURRENT_SETTINGS) == NULL) || ((CURRENT_SETTINGS)->log_level >= LOG_LEVEL_ERROR)) spamdyke_log(CURRENT_SETTINGS,LOG_LEVEL_ERROR,__VA_ARGS__); })
#define SPAMDYKE_LOG_INFO(CURRENT_SETTINGS,...)         ({ if (((CURRENT_SETTINGS) == NULL) || ((CURRENT_SETTINGS)->log_level >= LOG_LEVEL_INFO)) spamdyke_log(CURRENT_SETTINGS,LOG_LEVEL_INFO,__VA_ARGS__); })
#define SPAMDYKE_LOG_DEBUG(CURRENT_SETTINGS,...)        ({ if (((CURRENT_SETTINGS) == NULL) || ((CURRENT_SETTINGS)->log_level >= LOG_LEVEL_DEBUG)) spamdyke_log(CURRENT_SETTINGS,LOG_LEVEL_DEBUG,__VA_ARGS__); })
#define SPAMDYKE_LOG_EXCESSIVE(CURRENT_SETTINGS,...)    ({ if (((CURRENT_SETTINGS) == NULL) || ((CURRENT_SETTINGS)->log_level >= LOG_LEVEL_EXCESSIVE)) spamdyke_log(CURRENT_SETTINGS,LOG_LEVEL_EXCESSIVE,__VA_ARGS__); })

#define SYSLOG_IDENTIFIER               "spamdyke"
#define LOG_MISSING_DATA                "(unknown)"
#define LOG_ERROR_GRAYLIST_FILE         "ERROR: cannot write to graylist file "
#define LOG_ERROR_MOVE_DESCRIPTORS      "ERROR: unable to move file descriptors: "
#define LOG_ERROR_FORK                  "ERROR: unable to fork: "
#define LOG_ERROR_PIPE                  "ERROR: unable to create pipe: "
#define LOG_ERROR_EXEC                  "ERROR: unable to execute child process "
#define LOG_ERROR_EXEC_FILE             "ERROR: unable to find executable "
#define LOG_ERROR_OPEN                  "ERROR: unable to open file "
#define LOG_ERROR_OPEN_KEYWORDS         "ERROR: unable to open keywords file "
#define LOG_ERROR_OPEN_SEARCH           "ERROR: unable to open search file "
#define LOG_ERROR_RES_INIT              "ERROR: unable to initialize resolver library."
#define LOG_ERROR_GETPROTOBYNAME        "ERROR: unable to find protocol number with getprotobyname()\n"
#define LOG_ERROR_SOCKET                "ERROR: unable to create socket: %s"
#define LOG_ERROR_BIND                  "ERROR: unable to bind socket: %s"
#define LOG_ERROR_SETSOCKOPT            "ERROR: unable to set socket option: %s"
#define LOG_ERROR_SENDTO_INCOMPLETE     "ERROR: unable to send complete data packet, tried to send %d bytes, actually sent %d bytes"
#define LOG_ERROR_SENDTO                "ERROR: unable to send data packet: %s"
#define LOG_ERROR_MKQUERY               "ERROR: unable to construct data packet: %s"
#define LOG_ERROR_DN_EXPAND             "ERROR: dn_expand failed for %s; this could indicate a problem with the nameserver."
#define LOG_ERROR_DN_SKIPNAME           "ERROR: dn_skipname failed for %s; this could indicate a problem with the nameserver."
#define LOG_ERROR_DNS_RESPONSE          "ERROR: bad or invalid dns response to %s; this could indicate a problem with the name server."
#define LOG_ERROR_DNS_UNKNOWN_TYPE      "ERROR: dns response for %s: expected type %s but received type %s"
#define LOG_ERROR_STAT                  "ERROR: unable to stat graylist folder "
#define LOG_ERROR_MKDIR                 "ERROR: unable to create graylist folder "
#define LOG_ERROR_OPEN_LOG              "ERROR: unable to open traffic log file "
#define LOG_ERROR_REALLOC               "ERROR: out of memory - unable to allocate %lu bytes"
#define LOG_ERROR_FILE_TOO_LONG         "ERROR: ignoring file content past line %d: "
#define LOG_ERROR_TLS_INIT              "ERROR: unable to initialize SSL/TLS library"
#define LOG_ERROR_TLS_CERTIFICATE       "ERROR: unable to load SSL/TLS certificate from file: "
#define LOG_ERROR_TLS_PRIVATEKEY        "ERROR: unable to load or decrypt SSL/TLS private key from file: "
#define LOG_ERROR_TLS_CERT_CHECK        "ERROR: incorrect SSL/TLS private key password or SSL/TLS certificate/privatekey mismatch"
#define LOG_ERROR_TLS_ACCEPT            "ERROR: unable to start SSL/TLS connection"
#define LOG_ERROR_TLS_NO_ERROR          "Operation failed but no error was reported by the SSL/TLS library."
#define LOG_ERROR_TLS_ZERO_RETURN       "The connection was unexpectedly ended/closed"
#define LOG_ERROR_TLS_RECALL            "The SSL/TLS library wants a function to be called again to complete. This shouldn't happen"
#define LOG_ERROR_TLS_SYSCALL           "The operation failed due to an I/O error"
#define LOG_ERROR_TLS_LIBRARY           "A protocol or library failure occurred"
#define LOG_ERROR_TLS_EOF_FOUND         "Unexpected EOF found"
#define LOG_ERROR_AUTH_SUCCESS          "INFO: authentication successful: "
#define LOG_ERROR_AUTH_FAILURE          "ERROR: authentication failure (bad username/password, vchkpw uses this to indicate SMTP access is not allowed): "
#define LOG_ERROR_AUTH_MISUSE           "ERROR: authentication misuse (no input given or no additional command path given, e.g. /bin/true): "
#define LOG_ERROR_AUTH_ERROR            "ERROR: authentication error (likely due to missing/unexecutable commands): "
#define LOG_ERROR_AUTH_VCHKPW_BAD_CHARS "ERROR: authentication error %d (vchkpw uses this to indicate an unparsable email address): "
#define LOG_ERROR_AUTH_VCHKPW_UNKNOWN_USER      "ERROR: authentication error %d (vchkpw uses this to indicate an unknown username or authentication failure): "
#define LOG_ERROR_AUTH_VCHKPW_ENV_USER  "ERROR: authentication error %d (vchkpw uses this to indicate a failure to set the USER environment variable, possibly due to low memory): "
#define LOG_ERROR_AUTH_VCHKPW_ENV_HOME  "ERROR: authentication error %d (vchkpw uses this to indicate a failure to set the HOME environment variable, possibly due to low memory): "
#define LOG_ERROR_AUTH_VCHKPW_ENV_SHELL "ERROR: authentication error %d (vchkpw uses this to indicate a failure to set the SHELL environment variable, possibly due to low memory): "
#define LOG_ERROR_AUTH_VCHKPW_ENV_VPOPUSER      "ERROR: authentication error %d (vchkpw uses this to indicate a failure to set the VPOPUSER environment variable, possibly due to low memory): "
#define LOG_ERROR_AUTH_VCHKPW_BAD_INPUT "ERROR: authentication error %d (vchkpw uses this to indicate missing input on file descriptor 3): "
#define LOG_ERROR_AUTH_VCHKPW_NULL_USER "ERROR: authentication error %d (vchkpw uses this to indicate an empty username): "
#define LOG_ERROR_AUTH_VCHKPW_NULL_PASSWORD     "ERROR: authentication error %d (vchkpw uses this to indicate an empty password): "
#define LOG_ERROR_AUTH_VCHKPW_HOME_DIR  "ERROR: authentication error %d (vchkpw uses this to indicate a failure to create a virtual user's home directory): "
#define LOG_ERROR_AUTH_VCHKPW_NO_PASSWORD       "ERROR: authentication error %d (vchkpw uses this to indicate a virtual user has no password): "
#define LOG_ERROR_AUTH_VCHKPW_UNKNOWN_SYSTEM_USER       "ERROR: authentication error %d (vchkpw uses this to indicate a failed username lookup in the system password file): "
#define LOG_ERROR_AUTH_VCHKPW_UNKNOWN_SYSTEM_SHADOW     "ERROR: authentication error %d (vchkpw uses this to indicate a failed username lookup in the system shadow file): "
#define LOG_ERROR_AUTH_VCHKPW_FAILURE_SYSTEM_USER       "ERROR: authentication error %d (vchkpw uses this to indicate a failed authentication attempt using the system password file): "
#define LOG_ERROR_AUTH_UNKNOWN          "ERROR: unknown authentication error code "
#define LOG_ERROR_AUTH_ABEND            "ERROR: authentication aborted abnormally: "
#define LOG_ERROR_WRITE                 "ERROR: unable to write %d bytes to file descriptor %d: "
#define LOG_ERROR_BAD_CONFIG            "ERROR: unparsable configuration option in file %s on line %d: %s"
#define LOG_ERROR_OPEN_CONFIG           "ERROR: unable to open config file %s: %s"
#define LOG_ERROR_COMMAND_ABEND         "WARNING: command aborted abnormally: "

#define LOG_INFO_DNS_QUERY              "INFO: querying %s with DNS server %s:%d (attempt %d)"
#define LOG_INFO_DNS_RECEIVED           "INFO: received DNS packet: %d bytes"
#define LOG_INFO_DNS_RECEIVED_TYPE      "INFO: received DNS response: %s"
#define LOG_INFO_EXEC                   "INFO: executing command: %s"
#define LOG_INFO_EXEC_CHECKPASSWORD     "INFO: executing SMTP AUTH command %s for user: %s"
#define LOG_INFO_DNS_TXT                "INFO: found TXT record for %s: %.*s"
#define LOG_INFO_DNS_A                  "INFO: found A record for %s: %d.%d.%d.%d"
#define LOG_INFO_DNS_CNAME              "INFO: found CNAME record for %s: %s"
#define LOG_INFO_DNS_PTR                "INFO: found PTR record for %s: %s"
#define LOG_INFO_DNS_MX                 "INFO: found MX record for %s: %d %s"
#define LOG_INFO_DNS_NEGATIVE           "INFO: found no records for %s"
#define LOG_INFO_DOMAIN_DIR             "INFO: searching for domain directory entry: %s"

#define ERROR_CONFIG_NO_COMMAND         "ERROR: Missing qmail-smtpd command\n"
#define ERROR_CONFIG_UNKNOWN_OPTION     "ERROR: Unknown or incomplete option: %s\n"
#define ERROR_CONFIG_UNKNOWN_OPTION_FILE        "ERROR: Unknown configuration file option in file %s on line %d: %s\n"
#define ERROR_CONFIG_BAD_VALUE          "ERROR: Bad or unparsable value for option %s: %s\n"
#define ERROR_CONFIG_BAD_INTEGER_RANGE  "ERROR: Illegal value for option %s: %s (must be between %d and %d)\n"
#define ERROR_CONFIG_ILLEGAL_OPTION_CMDLINE     "ERROR: Option not allowed on command line: %s\n"
#define ERROR_CONFIG_ILLEGAL_OPTION_FILE        "ERROR: Option not allowed in configuration file, found in file %s on line %d: %s\n"

#define LOG_ACTION_TLS_PASSTHROUGH_START        -6
#define LOG_ACTION_AUTH_FAILURE                 -5
#define LOG_ACTION_AUTH_SUCCESS                 -4
#define LOG_ACTION_TLS_START                    -3
#define LOG_ACTION_TLS_END                      -2
#define LOG_ACTION_NONE                         -1
#define LOG_ACTION_REMOTE_FROM                  0
#define LOG_ACTION_CHILD_FROM                   1
#define LOG_ACTION_CHILD_FROM_DISCARDED         2
#define LOG_ACTION_FILTER_FROM                  3
#define LOG_ACTION_FILTER_TO                    4
#define LOG_ACTION_PREFIX                       (char *[]){ ">>> ", "<<< ", "<XX ", "<FF ", "FF>" }
#define LOG_ACTION_PREFIX_NONE                  ""
#define LOG_ACTION_PREFIX_TLS_SPAMDYKE          " TLS"
#define LOG_ACTION_PREFIX_TLS_PASSTHROUGH       " TLS_PASSTHROUGH"
#define LOG_ACTION_PREFIX_AUTH                  " AUTH:"
#define LOG_MESSAGE_TLS_PASSTHROUGH_START       "TLS passthrough started"
#define LOG_MESSAGE_TLS_START                   "TLS negotiated and started"
#define LOG_MESSAGE_TLS_END                     "TLS ended and closed"
#define LOG_MESSAGE_AUTH_SUCCESS                "Authentication successful"
#define LOG_MESSAGE_AUTH_FAILURE                "Authentication failed: "

#define LOG_MESSAGE_DNS_SEPARATOR               ", "
#define LOG_MESSAGE_DNS_TYPE_A                  "A"
#define LOG_MESSAGE_DNS_TYPE_CNAME              "CNAME"
#define LOG_MESSAGE_DNS_TYPE_MX                 "MX"
#define LOG_MESSAGE_DNS_TYPE_NS                 "NS"
#define LOG_MESSAGE_DNS_TYPE_PTR                "PTR"
#define LOG_MESSAGE_DNS_TYPE_SOA                "SOA"
#define LOG_MESSAGE_DNS_TYPE_TXT                "TXT"

#define CONFIG_TEST_ENVIRONMENT_LOCAL_PORT      "TCPLOCALPORT=25"
#define CONFIG_TEST_ENVIRONMENT_REMOTE_IP       "TCPREMOTEIP=127.0.0.1"
#define CONFIG_TEST_ENVIRONMENT_REMOTE_NAME     "TCPREMOTEHOST=localhost"

#define CONFIG_TEST_BAD_CONFIG_DIR_EXEC "ERROR(%s): Impossible test condition (DIR_EXEC). Please report this error to the author."
#define CONFIG_TEST_BAD_CONFIG_CMD_READ "ERROR(%s): Impossible test condition (CMD_READ). Please report this error to the author."
#define CONFIG_TEST_BAD_CONFIG_CMD_WRITE        "ERROR(%s): Impossible test condition (CMD_WRITE). Please report this error to the author."
#define CONFIG_TEST_BAD_CONFIG_CMD_READ_WRITE   "ERROR(%s): Impossible test condition (CMD_READ_WRITE). Please report this error to the author."

#define CONFIG_TEST_START               "Testing configuration...\n"
#define CONFIG_TEST_SUCCESS             "SUCCESS: Tests complete. No errors detected."
#define CONFIG_TEST_ERROR               "ERROR: Tests complete. Errors detected."

#define CONFIG_TEST_FAILURE_GETUSER     "ERROR: Unable to find user with name or ID %s"
#define CONFIG_TEST_SUCCESS_SETUSER     "SUCCESS: Set current user to %s (%d)."
#define CONFIG_TEST_FAILURE_SETUSER     "FAILURE: Unable to set current user to %s (%d): %s"
#define CONFIG_TEST_FAILURE_GETGROUP    "ERROR: Unable to find group with name or ID %s"
#define CONFIG_TEST_SUCCESS_SETGROUP    "SUCCESS: Set current group to %s (%d)."
#define CONFIG_TEST_FAILURE_SETGROUP    "FAILURE: Unable to set current group to %s (%d): %s"
#define CONFIG_TEST_SUCCESS_UID         "WARNING: Running tests as user %s (%d), group %s (%d). Is this the same user and group the mail server uses?"
#define CONFIG_TEST_ERROR_UID           "WARNING: Running tests as superuser %s (%d), group %s (%d). These test results may not be valid if the mail server runs as another user."

#define CONFIG_TEST_SUCCESS_SETUID      "SUCCESS: spamdyke binary (%s) is not owned by root and/or is not marked setuid."
#define CONFIG_TEST_ERROR_SETUID        "ERROR: spamdyke binary (%s) is owned by root and marked setuid. This is not necessary or recommended; it could be a security hole if exploitable bugs exist in spamdyke."
#define CONFIG_TEST_ERROR_SETUID_STAT   "ERROR: Unable to stat() spamdyke binary (%s) to scan permissions: %s"
#define CONFIG_TEST_ERROR_SETUID_SEARCH "ERROR: Unable to find spamdyke binary (%s) in the current path."
#define CONFIG_TEST_ERROR_SETUID_FILENAME       "ERROR: Name of current binary is unknown. This condition should be impossible."

#define CONFIG_TEST_START_FILE_READ     "INFO(%s): Testing file read: %s"
#define CONFIG_TEST_SUCCESS_FILE_READ   "SUCCESS(%s): Opened for reading: %s"
#define CONFIG_TEST_ERROR_FILE_READ     "ERROR(%s): Failed to open for reading: %s: %s"
#define CONFIG_TEST_START_FILE_WRITE    "INFO(%s): Testing file write: %s"
#define CONFIG_TEST_SUCCESS_FILE_WRITE  "SUCCESS(%s): Opened for writing: %s"
#define CONFIG_TEST_ERROR_FILE_WRITE    "ERROR(%s): Failed to open for writing: %s: %s"
#define CONFIG_TEST_START_FILE_READ_WRITE       "INFO(%s): Testing file reading and writing: %s"
#define CONFIG_TEST_SUCCESS_FILE_READ_WRITE     "SUCCESS(%s): Opened for reading and writing: %s"
#define CONFIG_TEST_ERROR_FILE_READ_WRITE       "ERROR(%s): Failed to open for reading and writing: %s: %s"
#define CONFIG_TEST_START_EXECUTE       "INFO(%s): Testing executable: %s"
#define CONFIG_TEST_SUCCESS_EXECUTE     "SUCCESS(%s): File is executable: %s"
#define CONFIG_TEST_ERROR_EXECUTE       "ERROR(%s): File is not executable: %s: %s"
#define CONFIG_TEST_START_DIR_READ      "INFO(%s): Testing directory tree and all files for reading: %s"
#define CONFIG_TEST_SUCCESS_DIR_READ    "SUCCESS(%s): Directory tree and all files are readable: %s"
#define CONFIG_TEST_ERROR_DIR_READ      "ERROR(%s): Portions of directory tree and/or some files are not readable: %s: %s"
#define CONFIG_TEST_START_DIR_WRITE     "INFO(%s): Testing directory for writing: %s"
#define CONFIG_TEST_SUCCESS_DIR_WRITE   "SUCCESS(%s): Created and deleted file in directory: %s"
#define CONFIG_TEST_ERROR_DIR_WRITE     "ERROR(%s): Failed to create file in directory: %s: %s"
#define CONFIG_TEST_ERROR_DIR_WRITE_DELETE      "ERROR(%s): Failed to delete test file in directory: %s: %s"

#define CONFIG_TEST_START_TLS           "INFO(%s): Testing TLS by initializing SSL/TLS library with certificate and key"
#define CONFIG_TEST_START_TLS_PRIVATEKEY        "INFO(%s): Testing TLS private key file for reading: %s"
#define CONFIG_TEST_SUCCESS_TLS         "SUCCESS(%s): Certificate and key loaded; SSL/TLS library successfully initialized"
#define CONFIG_TEST_ERROR_TLS_CERT_DISABLED             "ERROR(%s): TLS support is not compiled into this executable but a TLS certificate file was given anyway: %s"
#define CONFIG_TEST_ERROR_TLS_PRIVATEKEY_DISABLED       "ERROR(%s): TLS support is not compiled into this executable but a TLS private key file was given anyway: %s"
#define CONFIG_TEST_ERROR_TLS_PASSWORD_DISABLED         "ERROR(%s): TLS support is not compiled into this executable but a TLS private key password was given anyway."

#define CONFIG_TEST_SMTPAUTH_START      "INFO(%s): Examining authentication command: %s"
#define CONFIG_TEST_SMTPAUTH_OWNER_WARN "WARNING(%s): Authentication command is not owned by root. Some require being setuid root to read system passwords: %s: owned by %s (%d)"
#define CONFIG_TEST_SMTPAUTH_SETUID_WARN        "WARNING(%s): Authentication command is owned by root but not setuid. Some require being setuid root to read system passwords: %s"
#define CONFIG_TEST_SMTPAUTH_RUN_PLAIN  "INFO(%s): Running authentication command with unencrypted input: %s"
#define CONFIG_TEST_SUCCESS_SMTPAUTH_PLAIN      "SUCCESS(%s): Authentication succeeded with unencrypted input: %s"
#define CONFIG_TEST_FAILURE_SMTPAUTH_PLAIN      "ERROR(%s): Authentication failed with unencrypted input: %s"
#define CONFIG_TEST_SMTPAUTH_RUN_ENCRYPTED      "INFO(%s): Running authentication command with encrypted input: %s"
#define CONFIG_TEST_SUCCESS_SMTPAUTH_ENCRYPTED  "SUCCESS(%s): Authentication succeeded with encrypted input: %s"
#define CONFIG_TEST_FAILURE_SMTPAUTH_ENCRYPTED  "ERROR(%s): Authentication failed with encrypted input: %s"
#define CONFIG_TEST_SMTPAUTH_SUGGEST_ENCRYPTED  "INFO: One or more authentication commands support encrypted input; use \"smtp-auth-command-encryption\" instead of \"smtp-auth-command\""
#define CONFIG_TEST_SMTPAUTH_SUGGEST_PLAIN      "INFO: No authentication commands support encrypted input; use \"smtp-auth-command\" instead of \"smtp-auth-command-encryption\""

#define CONFIG_TEST_TYPE_IFIFO          "FIFO (i.e. a named pipe)"
#define CONFIG_TEST_TYPE_IFCHR          "character device (e.g. a serial port)"
#define CONFIG_TEST_TYPE_IFDIR          "directory"
#define CONFIG_TEST_TYPE_IFBLK          "block device (e.g. a hard disk)"
#define CONFIG_TEST_TYPE_IFREG          "regular file"
#define CONFIG_TEST_TYPE_IFLNK          "symbolic link"
#define CONFIG_TEST_TYPE_IFSOCK         "socket"
#define CONFIG_TEST_TYPE_IFWHT          "whiteout"
#define CONFIG_TEST_TYPE_UNKNOWN        "filesystem entry of unknown type"

#define CONFIG_TEST_MSG_REGULAR_FILE    "File is not a regular file, it is a(n) %s"
#define CONFIG_TEST_MSG_OWNER_NO_EXEC   "Owner permissions apply but owner executable bit is not set"
#define CONFIG_TEST_MSG_PGROUP_NO_EXEC  "Group permissions apply but group executable bit is not set"
#define CONFIG_TEST_MSG_SGROUP_NO_EXEC  "Group permissions apply (through a secondary group) but group executable bit is not set"
#define CONFIG_TEST_MSG_OTHER_NO_EXEC   "\"Other\" permissions apply but \"other\" executable bit is not set"

#define CONFIG_TEST_START_GRAYLIST      "INFO(%s): Testing graylist directory: %s"
#define CONFIG_TEST_SUCCESS_GRAYLIST    "SUCCESS(%s): Graylist directory tests succeeded: %s"
#define CONFIG_TEST_ERROR_GRAYLIST_TOP_EMPTY    "ERROR(%s): No domain directories found in graylist folder: %s"
#define CONFIG_TEST_ERROR_GRAYLIST_TOP_DIR      "ERROR(%s): Unable to read graylist directory %s: %s"
#define CONFIG_TEST_ERROR_GRAYLIST_TOP_OTHER    "ERROR(%s): Found %s in graylist folder where only domain directories should be: %s"
#define CONFIG_TEST_ERROR_GRAYLIST_DOMAIN_DIR   "ERROR(%s): Unable to read graylist domain directory %s: %s"
#define CONFIG_TEST_ERROR_GRAYLIST_DOMAIN_OTHER "ERROR(%s): Found %s in graylist domain folder where only user directories should be: %s"
#define CONFIG_TEST_ERROR_GRAYLIST_USER_DIR     "ERROR(%s): Unable to read graylist user directory %s: %s"
#define CONFIG_TEST_ERROR_GRAYLIST_USER_OTHER   "ERROR(%s): Found %s in graylist user folder where only user directories should be: %s"

#define CONFIG_TEST_START_RDNS_DIR      "INFO(%s): Testing rDNS directory: %s"
#define CONFIG_TEST_SUCCESS_RDNS_DIR    "SUCCESS(%s): rDNS directory tests succeeded: %s"
#define CONFIG_TEST_ERROR_RDNS_LETTER_DIR       "ERROR(%s): rDNS directory name is longer than one character: %s"
#define CONFIG_TEST_ERROR_RDNS_LETTER_MISMATCH  "ERROR(%s): rDNS directory name does not start with the same single character as its parent: %s"
#define CONFIG_TEST_ERROR_RDNS_FQDN_MISMATCH    "ERROR(%s): rDNS entry does not match domain name %s: %s"
#define CONFIG_TEST_ERROR_RDNS_NO_FILES "ERROR(%s): rDNS directory contains no files: %s"
#define CONFIG_TEST_ERROR_RDNS_NO_FOLDERS       "ERROR(%s): rDNS directory contains no subdirectories: %s"
#define CONFIG_TEST_ERROR_RDNS_OPENDIR  "ERROR(%s): Unable to read rDNS directory: %s"
#define CONFIG_TEST_ERROR_RDNS_NON_DIR  "ERROR(%s): Found %s in rDNS directory where only directories should be: %s"
#define CONFIG_TEST_ERROR_RDNS_NON_FILE "ERROR(%s): Found %s in rDNS directory where only files should be: %s"

#define CONFIG_TEST_PATCH_SUCCESS_CONTINUATION  "\r\n250-"
#define CONFIG_TEST_PATCH_SUCCESS_END           "\r\n250 "
#define CONFIG_TEST_PATCH_TLS                   "starttls\r\n"
#define CONFIG_TEST_PATCH_SMTP_AUTH             "auth "
#define CONFIG_TEST_PATCH_EXPECT_GREETING       "220 "
#define CONFIG_TEST_PATCH_EXPECT_EHLO           "\r\n250 "
#define CONFIG_TEST_PATCH_SEND_EHLO             "EHLO localhost\r\n"
#define CONFIG_TEST_PATCH_SEND_QUIT             "QUIT\r\n"
#define CONFIG_TEST_PATCH_SCRIPT                { \
                                                  { ES_TYPE_EXPECT, CONFIG_TEST_PATCH_EXPECT_GREETING, STRLEN(CONFIG_TEST_PATCH_EXPECT_GREETING) }, \
                                                  { ES_TYPE_SEND, CONFIG_TEST_PATCH_SEND_EHLO, STRLEN(CONFIG_TEST_PATCH_SEND_EHLO) }, \
                                                  { ES_TYPE_EXPECT, CONFIG_TEST_PATCH_EXPECT_EHLO, STRLEN(CONFIG_TEST_PATCH_EXPECT_EHLO) }, \
                                                  { ES_TYPE_SEND, CONFIG_TEST_PATCH_SEND_QUIT, STRLEN(CONFIG_TEST_PATCH_SEND_QUIT) }, \
                                                  { ES_TYPE_NONE, NULL, 0 } \
                                                }
#define CONFIG_TEST_PATCH_RUN                   "INFO: Running command to test capabilities: %s"
#define CONFIG_TEST_ERROR_PATCH_NO_OUTPUT       "ERROR: Command returned no output: %s"
#define CONFIG_TEST_SUCCESS_PATCH_TLS           "SUCCESS: %s appears to offer TLS support. Continue using the \"tls-certificate-file\" flag so spamdyke will be able to filter all traffic."
#define CONFIG_TEST_SUCCESS_PATCH_TLS_FLAG      "WARNING: %s appears to offer TLS support but spamdyke cannot offer TLS without the \"tls-certificate-file\" flag. Please use \"tls-certificate-file\" so spamdyke can filter all traffic. The certificate is typically stored in the /var/qmail/control directory."
#define CONFIG_TEST_ERROR_PATCH_TLS             "SUCCESS: %s does not appear to offer TLS support. Continue using the \"tls-certificate-file\" flag so spamdyke will be able to filter all traffic."
#define CONFIG_TEST_ERROR_PATCH_TLS_FLAG        "WARNING: %s does not appear to offer TLS support. Please generate a certificate and use the \"tls-certificate-file\" flag so spamdyke can offer TLS."
#define CONFIG_TEST_SUCCESS_PATCH_SMTP_AUTH     "SUCCESS: %s appears to offer SMTP AUTH support. spamdyke will observe any authentication and trust its response."
#define CONFIG_TEST_SUCCESS_PATCH_SMTP_AUTH_FLAG        "WARNING: %s appears to offer SMTP AUTH support but the \"smtp-auth-command\", \"smtp-auth-command-encryption\" and/or \"access-file\" flags are in use. This is not necessary and needlessly creates extra load on the server."
#define CONFIG_TEST_ERROR_PATCH_SMTP_AUTH       "SUCCESS: %s does not appear to offer SMTP AUTH support. Continue using the \"smtp-auth-command\" flag or the \"smtp-auth-command-encryption\" flag as well as the \"access-file\" and \"local-domains-file\" flags so spamdyke will be able to authenticate users and correctly allow them to relay."
#define CONFIG_TEST_ERROR_PATCH_SMTP_AUTH_FLAG  "WARNING: %s does not appear to offer SMTP AUTH support. Please use the \"smtp-auth-command\" flag or the \"smtp-auth-command-encryption\" flag as well as the \"access-file\" and \"local-domains-file\" flags so spamdyke will be able to authenticate users and correctly allow them to relay."

#define CONFIG_TYPE_NONE                        -3
#define CONFIG_TYPE_ACTION_ONCE                 -2
#define CONFIG_TYPE_ACTION_MULTIPLE             -1
#define CONFIG_TYPE_BOOLEAN                     0
#define CONFIG_TYPE_INTEGER                     1
#define CONFIG_TYPE_STRING_SINGLETON            2
#define CONFIG_TYPE_STRING_ARRAY                3
#define CONFIG_TYPE_FILE_SINGLETON              4
#define CONFIG_TYPE_FILE_ARRAY                  5
#define CONFIG_TYPE_DIR_SINGLETON               6
#define CONFIG_TYPE_DIR_ARRAY                   7
#define CONFIG_TYPE_COMMAND_SINGLETON           8
#define CONFIG_TYPE_COMMAND_ARRAY               9

#define CONFIG_ACCESS_NONE                      0
#define CONFIG_ACCESS_READ_ONLY                 1
#define CONFIG_ACCESS_WRITE_ONLY                2
#define CONFIG_ACCESS_READ_WRITE                3
#define CONFIG_ACCESS_EXECUTE                   4

#define CONFIG_LOCATION_CMDLINE                 0x01
#define CONFIG_LOCATION_FILE                    0x02

#define CONFIG_ACTION(CMD)                      ({ int action(struct filter_settings *current_settings, int current_return_value, char *input_value, struct previous_action *history) { CMD; return(current_return_value); } &action; })
#define CONFIG_ACCESSOR_INTEGER(MEMBER)         ({ int *access_integer(struct filter_settings *current_settings) { return(&current_settings->MEMBER); } &access_integer; })
#define CONFIG_ACCESSOR_STRING(MEMBER)          ({ char **access_string(struct filter_settings *current_settings) { return(&current_settings->MEMBER); } &access_string; })
#define CONFIG_ACCESSOR_STRING_ARRAY(MEMBER)    ({ char ***access_string_array(struct filter_settings *current_settings) { return(&current_settings->MEMBER); } &access_string_array; })

#define ES_TYPE_NONE                            0
#define ES_TYPE_SEND                            1
#define ES_TYPE_EXPECT                          2

struct expect_send
  {
  int type;
  char *data;
  int strlen_data;
  };

struct filter_settings;

struct previous_action
  {
  char *data;
  int count;
  struct previous_action *prev;
  };

struct spamdyke_option
  {
  int value_type;
  int access_type;
  int location;
  struct option getopt_option;
  union
    {
    int integer_value;
    char *string_value;
    } default_value;
  union
    {
    int integer_value;
    char *string_value;
    } missing_value;
  union
    {
    int *(*get_integer)(struct filter_settings *);
    char **(*get_string)(struct filter_settings *);
    char ***(*get_string_array)(struct filter_settings *);
    } getter;
  int integer_minimum;
  int integer_maximum;
  int set_consequence;
  int set_grace;
  int (*test_function)(struct filter_settings *, struct spamdyke_option *, char **);
  int (*additional_set_actions)(struct filter_settings *, int, char *, struct previous_action *);
  char *help_argument;
  char *help_text;
  };

struct filter_settings
  {
  struct spamdyke_option *option_list;
  int num_options;
  struct option *long_options;
  char short_options[MAX_BUF + 1];
  struct spamdyke_option **option_lookup;
  int max_short_code;

  char server_name[MAX_BUF + 1];
  int strlen_server_name;
  char *server_ip;
  int strlen_server_ip;

  char *local_server_name;
  char *local_server_name_file;
  char *local_server_name_command;

  char **graylist_dir;
  int graylist_min_secs;
  int graylist_max_secs;
  int max_rcpt_to;
  int log_target;
  int log_level;
  char *policy_location;
  int strlen_policy_location;
  char *log_dir;
  char **blacklist_sender_file;
  char **whitelist_sender_file;
  char **blacklist_recipient_file;
  char **whitelist_recipient_file;

  char **access_list_file;
  char **local_domains_file;
  char **blacklist_rdns_keyword_file;
  char **blacklist_rdns_file;
  char **blacklist_rdns_dir;
  char **blacklist_ip_file;
  char **whitelist_rdns_file;
  char **whitelist_rdns_dir;
  char **whitelist_ip_file;
  char **dnsrwl_fqdn;
  char **dnsrbl_fqdn;
  char **rhswl_fqdn;
  char **rhsbl_fqdn;
  char **graylist_never_ip_file;
  char **graylist_never_rdns_file;
  char **graylist_never_rdns_dir;
  char **graylist_always_ip_file;
  char **graylist_always_rdns_file;
  char **graylist_always_rdns_dir;
  char **smtp_auth_command;
  int smtp_auth_encryption;

  int graylist_type;
  int check_ip_in_rdns_cc;
  int check_earlytalker;
  int check_rdns_exist;
  int check_rdns_resolve;
  int check_sender_mx;

  int timeout_connection;
  int timeout_command;

  int allow_relay;
  int inside_data;
  int num_rcpt_to;
  int local_sender;

  char sender_address[MAX_ADDRESS + 1];
  char recipient_address[MAX_ADDRESS + 1];
  char graylist_path[MAX_PATH + 1];

  struct rejection_data *rejection;
  struct rejection_data *transient_rejection;
  struct rejection_data rejection_buf;
  struct rejection_data transient_rejection_buf;
  char reject_message_buf[MAX_BUF + 1];
  char short_reject_message_buf[MAX_BUF + 1];
  char transient_reject_message_buf[MAX_BUF + 1];
  char transient_short_reject_message_buf[MAX_BUF + 1];

  char **child_argv;
  int filter_action;
  int filter_grace;

  int smtp_auth_state;
  int smtp_auth_type;
  int smtp_auth_origin;
  char smtp_auth_challenge[MAX_BUF + 1];
  char smtp_auth_response[MAX_BUF + 1];
  char smtp_auth_username[MAX_BUF + 1];

  int tls_state;
  char *tls_certificate_file;
  char *tls_privatekey_file;
  int tls_strlen_privatekey_password;
  char *tls_privatekey_password;
  char *tls_privatekey_password_file;

  char *test_smtp_auth_username;
  char *test_smtp_auth_password;
  char *test_user;

#ifdef HAVE_LIBSSL

  SSL_CTX *tls_context;
  SSL *tls_session;

#endif /* HAVE_LIBSSL */

  };

extern struct rejection_data *rejection;

#endif /* SPAMDYKE_H */
