/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "spamdyke.h"
#include "configuration.h"
#include "usage.h"

void print_help_text(char *input_text)
  {
  int strlen_text;
  int cur_loc;
  int next_loc;

  if (input_text != NULL)
    {
    strlen_text = strlen(input_text);
    for (cur_loc = 0; (cur_loc < strlen_text) && (input_text[cur_loc] == ' '); cur_loc++);

    while (cur_loc < strlen_text)
      {
      if ((strlen_text - cur_loc) > USAGE_LINE_WRAP)
        {
        for (next_loc = cur_loc + (USAGE_LINE_WRAP - STRLEN(USAGE_LINE_INDENT)) + 1; (next_loc > cur_loc) && (input_text[next_loc] != ' '); next_loc--);
        if (next_loc == cur_loc)
          for (next_loc = cur_loc + (USAGE_LINE_WRAP - STRLEN(USAGE_LINE_INDENT)) + 1; (next_loc < strlen_text) && (input_text[next_loc] != ' '); next_loc++);
        }
      else
        next_loc = strlen_text;

      fprintf(stderr, USAGE_LINE_INDENT "%.*s\n", next_loc - cur_loc, input_text + cur_loc);

      for (cur_loc = next_loc + 1; (cur_loc < strlen_text) && (input_text[cur_loc] == ' '); cur_loc++);
      }
    }

  return;
  }

void usage(struct filter_settings *current_settings, int full_text, char *error_pattern, ...)
  {
  static int banner_printed = 0;
  int i;
  va_list tmp_va;

  if (!banner_printed)
    {
    fprintf(stderr, USAGE_MESSAGE_HEADER);

    if (!full_text)
      fprintf(stderr, USAGE_MESSAGE_FOOTER_SHORT);

    banner_printed = 1;
    }

  if (error_pattern != NULL)
    {
    va_start(tmp_va, error_pattern);
    vprintf(error_pattern, tmp_va);
    va_end(tmp_va);
    }
  else if (full_text)
    {
    fprintf(stderr, USAGE_MESSAGE_USAGE);

    for (i = 0; current_settings->option_list[i].value_type != CONFIG_TYPE_NONE; i++)
      if (current_settings->option_list[i].help_text != NULL)
        {
        fprintf(stderr, "\n");

        if (current_settings->option_list[i].getopt_option.val <= current_settings->max_short_code)
          {
          if ((current_settings->option_list[i].getopt_option.has_arg == optional_argument) &&
              (current_settings->option_list[i].help_argument != NULL))
            fprintf(stderr, "-%c[%s]\n", current_settings->option_list[i].getopt_option.val, current_settings->option_list[i].help_argument);
          else if ((current_settings->option_list[i].getopt_option.has_arg == required_argument) &&
                   (current_settings->option_list[i].help_argument != NULL))
            fprintf(stderr, "-%c %s\n", current_settings->option_list[i].getopt_option.val, current_settings->option_list[i].help_argument);
          else
            fprintf(stderr, "-%c\n", current_settings->option_list[i].getopt_option.val);
          }

        if (current_settings->option_list[i].getopt_option.name != NULL)
          {
          if ((current_settings->option_list[i].getopt_option.has_arg == optional_argument) &&
              (current_settings->option_list[i].help_argument != NULL))
            fprintf(stderr, "--%s=[%s]\n", current_settings->option_list[i].getopt_option.name, current_settings->option_list[i].help_argument);
          else if ((current_settings->option_list[i].getopt_option.has_arg == required_argument) &&
                   (current_settings->option_list[i].help_argument != NULL))
            fprintf(stderr, "--%s %s\n", current_settings->option_list[i].getopt_option.name, current_settings->option_list[i].help_argument);
          else
            fprintf(stderr, "--%s\n", current_settings->option_list[i].getopt_option.name);
          }

        print_help_text(current_settings->option_list[i].help_text);

        if (current_settings->option_list[i].help_argument != NULL)
          {
          if (current_settings->option_list[i].value_type == CONFIG_TYPE_INTEGER)
            fprintf(stderr, USAGE_LINE_INDENT USAGE_MESSAGE_INTEGER_RANGE, current_settings->option_list[i].help_argument, current_settings->option_list[i].integer_minimum, current_settings->option_list[i].integer_maximum);

          if (current_settings->option_list[i].getopt_option.has_arg == optional_argument)
            {
            if (current_settings->option_list[i].getopt_option.val <= current_settings->max_short_code)
              fprintf(stderr, USAGE_LINE_INDENT USAGE_MESSAGE_OPTIONAL_SHORT, current_settings->option_list[i].getopt_option.val, current_settings->option_list[i].help_argument);

            if (current_settings->option_list[i].getopt_option.name != NULL)
              fprintf(stderr, USAGE_LINE_INDENT USAGE_MESSAGE_OPTIONAL_LONG, current_settings->option_list[i].getopt_option.name, current_settings->option_list[i].help_argument);
            }
          }

        switch (current_settings->option_list[i].value_type)
          {
          case CONFIG_TYPE_STRING_ARRAY:
          case CONFIG_TYPE_FILE_ARRAY:
          case CONFIG_TYPE_DIR_ARRAY:
          case CONFIG_TYPE_COMMAND_ARRAY:
          case CONFIG_TYPE_ACTION_MULTIPLE:
            fprintf(stderr, USAGE_LINE_INDENT USAGE_MESSAGE_ARRAY, current_settings->option_list[i].getopt_option.name);
            break;
          case CONFIG_TYPE_STRING_SINGLETON:
          case CONFIG_TYPE_FILE_SINGLETON:
          case CONFIG_TYPE_DIR_SINGLETON:
          case CONFIG_TYPE_COMMAND_SINGLETON:
          case CONFIG_TYPE_ACTION_ONCE:
            fprintf(stderr, USAGE_LINE_INDENT USAGE_MESSAGE_SINGLETON, current_settings->option_list[i].getopt_option.name);
            break;
          default:
            break;
          }
        }
    
    fprintf(stderr, USAGE_MESSAGE_FOOTER_LONG);
    }

  return;
  }
